using System;
using System.Collections.Generic;
using System.Text;

using System.Management.Automation;

namespace Microsoft.PowerShell.CmdletManagement
{
    [Cmdlet(VerbsCommon.Remove, "PsSpecNoun", SupportsShouldProcess = true, DefaultParameterSetName="ById")]
    public class RemovePsSpecNounCommand : Cmdlet
    {
        // ById Parameter Set
        [Parameter(Mandatory = true, Position = 0, ValueFromPipelineByPropertyName = true, ParameterSetName = "ById")]
        public int Id
        {
            get
            {
                return _id;
            }
            set
            {
                _id = value;
                idSpecified = true;
            }
        }
        private int _id = -1;
        private bool idSpecified = false;

        // ByName Parameter Set
        [Parameter(Mandatory = true, Position = 0, ValueFromPipelineByPropertyName = true, ParameterSetName = "ByName")]
        [ValidateLength(0, 50)]
        public String Name
        {
            get
            {
                return _name;
            }
            set
            {
                _name = value;
            }
        }
        private string _name = null;

        [Parameter]
        public SwitchParameter Force
        {
            get
            {
                return _force;
            }
            set
            {
                _force = value;
            }
        }
        private SwitchParameter _force;

        [Parameter(Mandatory = false)]
        public String ServiceUri
        {
            get
            {
                return _serviceUri;
            }
            set
            {
                _serviceUri = value;
            }
        }
        private string _serviceUri = PsSpecCmdletUtils.Constants.ServiceUri;

        protected override void ProcessRecord()
        {
            // Remove multiple nouns if specified by name
            if (! idSpecified)
            {
                foreach (PsSpecNoun noun in GetNouns(Name))
                {
                    RemoveNoun(noun.Id);
                }
            }
            else
            {
                RemoveNoun(Id);
            }
        }

        private IEnumerable<PsSpecNoun> GetNouns(string name)
        {
            // Check that the noun name is valid
            if (!PsSpecCmdletUtils.IsValidNounName(name))
            {
                WriteError(new ErrorRecord(
                    new Exception(resources.InvalidNounName),
                    "INVALID_NAME",
                    ErrorCategory.InvalidArgument, name));
                yield break;
            }

            GetPsSpecNounCommand psSpecNounCommand = new GetPsSpecNounCommand();
            psSpecNounCommand.Name = name;
            psSpecNounCommand.ServiceUri = ServiceUri;

            System.Collections.ArrayList results = new System.Collections.ArrayList();
            foreach (Object result in psSpecNounCommand.Invoke())
            {
                results.Add(result);
            }

            // Check that the noun exists
            if (results.Count == 0)
            {
                string errorMessage = String.Format(resources.NounNotFound, name);

                WriteError(new ErrorRecord(
                    new Exception(errorMessage),
                    "NOUN_NOT_FOUND",
                    ErrorCategory.ObjectNotFound, name));
                yield break;
            }

            // Throw an error if they haven't specified enough to disambiguate
            if (results.Count > 1)
            {
                string errorMessage = resources.DuplicateNounFound;

                WriteError(new ErrorRecord(
                    new Exception(errorMessage),
                    "DUPLICATE_NOUN_FOUND",
                    ErrorCategory.InvalidData, name));
                yield break;
            }

            foreach (Object result in results)
            {
                yield return (PsSpecNoun)result;
            }
        }

        private void RemoveNoun(int id)
        {
            if (ShouldProcess(
                String.Format("{0}", ServiceUri),
                String.Format(resources.DeleteNounShouldProcess, id)))
            {
                PsSpecCmdletService.CmdletService service =
                    new PsSpecCmdletService.CmdletService();
                service.Url = ServiceUri;
                service.UseDefaultCredentials = true;

                if (_force || ShouldContinue(
                    String.Format(resources.ReallyWantToRemoveNounText, id),
                    resources.ReallyWantToRemoveNounTitle))
                {
                    try
                    {
                        service.RemoveNoun(id);
                    }
                    catch (System.Web.Services.Protocols.SoapException e)
                    {
                        WriteError(new ErrorRecord(e, "CMDLET_SERVICE_ERROR", ErrorCategory.InvalidOperation, Name));
                    }
                }
            }
        }
    }
}
